"""Utilities for anoncreds presentation exchange attachment."""

import json
import logging
import time
from typing import Dict, Optional, Tuple, Union

from ....anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ....anoncreds.models.credential_definition import CredDef
from ....anoncreds.models.revocation import RevRegDef
from ....anoncreds.models.schema import AnonCredsSchema
from ....anoncreds.models.utils import extract_non_revocation_intervals_from_proof_request
from ....anoncreds.registry import AnonCredsRegistry
from ....anoncreds.revocation import AnonCredsRevocation
from ....askar.profile_anon import AskarAnonCredsProfile
from ....core.error import BaseError
from ....core.profile import Profile
from ..v1_0.models.presentation_exchange import V10PresentationExchange
from ..v2_0.messages.pres_format import V20PresFormat
from ..v2_0.models.pres_exchange import V20PresExRecord

LOGGER = logging.getLogger(__name__)


class AnonCredsPresExchHandlerError(BaseError):
    """Base class for AnonCreds Presentation Exchange related errors."""


class AnonCredsPresExchHandler:
    """Base Presentation Exchange Handler."""

    def __init__(
        self,
        profile: Profile,
    ):
        """Initialize PresExchange Handler."""
        super().__init__()
        self._profile = profile
        self.holder = AnonCredsHolder(profile)

    def _extract_proof_request(self, pres_ex_record):
        if isinstance(pres_ex_record, V20PresExRecord):
            return pres_ex_record.pres_request.attachment(
                V20PresFormat.Format.ANONCREDS
            ) or pres_ex_record.pres_request.attachment(V20PresFormat.Format.INDY)
        elif isinstance(pres_ex_record, V10PresentationExchange):
            return pres_ex_record._presentation_request.ser

        raise TypeError(
            "pres_ex_record must be V10PresentationExchange or V20PresExRecord"
        )

    def _get_requested_referents(
        self,
        proof_request: dict,
        requested_credentials: dict,
        non_revoc_intervals: dict,
    ) -> dict:
        """Get requested referents for a proof request and requested credentials.

        Returns a dictionary that looks like:
        {
          "referent-0": {"cred_id": "0", "non_revoked": {"from": ..., "to": ...}},
          "referent-1": {"cred_id": "1", "non_revoked": {"from": ..., "to": ...}}
        }
        """

        requested_referents = {}
        attr_creds = requested_credentials.get("requested_attributes", {})
        req_attrs = proof_request.get("requested_attributes", {})
        for reft in attr_creds:
            requested_referents[reft] = {"cred_id": attr_creds[reft]["cred_id"]}
            if reft in req_attrs and reft in non_revoc_intervals:
                requested_referents[reft]["non_revoked"] = non_revoc_intervals[reft]

        pred_creds = requested_credentials.get("requested_predicates", {})
        req_preds = proof_request.get("requested_predicates", {})
        for reft in pred_creds:
            requested_referents[reft] = {"cred_id": pred_creds[reft]["cred_id"]}
            if reft in req_preds and reft in non_revoc_intervals:
                requested_referents[reft]["non_revoked"] = non_revoc_intervals[reft]
        return requested_referents

    async def _get_credentials(self, requested_referents: dict):
        """Extract mapping of presentation referents to credential ids."""
        credentials = {}
        for reft in requested_referents:
            credential_id = requested_referents[reft]["cred_id"]
            if credential_id not in credentials:
                credentials[credential_id] = json.loads(
                    await self.holder.get_credential(credential_id)
                )
        return credentials

    def _remove_superfluous_timestamps(self, requested_credentials, credentials):
        """Remove any timestamps that cannot correspond to non-revoc intervals."""
        for r in ("requested_attributes", "requested_predicates"):
            for reft, req_item in requested_credentials.get(r, {}).items():
                if not credentials[req_item["cred_id"]].get(
                    "rev_reg_id"
                ) and req_item.pop("timestamp", None):
                    LOGGER.info(
                        f"Removed superfluous timestamp from requested_credentials {r} "
                        f"{reft} for non-revocable credential {req_item['cred_id']}"
                    )

    async def _get_ledger_objects(
        self, credentials: dict
    ) -> Tuple[Dict[str, AnonCredsSchema], Dict[str, CredDef], Dict[str, RevRegDef]]:
        """Get all schemas, credential definitions, and revocation registries in use."""
        schemas = {}
        cred_defs = {}
        revocation_registries = {}

        for credential in credentials.values():
            schema_id = credential["schema_id"]
            anoncreds_registry = self._profile.inject(AnonCredsRegistry)
            if schema_id not in schemas:
                schemas[schema_id] = (
                    await anoncreds_registry.get_schema(self._profile, schema_id)
                ).schema
            cred_def_id = credential["cred_def_id"]
            if cred_def_id not in cred_defs:
                cred_defs[cred_def_id] = (
                    await anoncreds_registry.get_credential_definition(
                        self._profile, cred_def_id
                    )
                ).credential_definition
            if credential.get("rev_reg_id"):
                revocation_registry_id = credential["rev_reg_id"]
                if revocation_registry_id not in revocation_registries:
                    rev_reg = (
                        await anoncreds_registry.get_revocation_registry_definition(
                            self._profile, revocation_registry_id
                        )
                    ).revocation_registry
                    revocation_registries[revocation_registry_id] = rev_reg

        return schemas, cred_defs, revocation_registries

    async def _get_revocation_lists(self, requested_referents: dict, credentials: dict):
        """Get revocation lists.

        Get revocation lists with non-revocation interval defined in
        "non_revoked" of the presentation request or attributes
        """
        epoch_now = int(time.time())
        rev_lists = {}
        for precis in requested_referents.values():  # cred_id, non-revoc interval
            credential_id = precis["cred_id"]
            if not credentials[credential_id].get("rev_reg_id"):
                continue
            if "timestamp" in precis:
                continue
            rev_reg_id = credentials[credential_id]["rev_reg_id"]

            anoncreds_registry = self._profile.inject(AnonCredsRegistry)
            reft_non_revoc_interval = precis.get("non_revoked")
            if reft_non_revoc_interval:
                key = (
                    f"{rev_reg_id}_"
                    f"{reft_non_revoc_interval.get('from', 0)}_"
                    f"{reft_non_revoc_interval.get('to', epoch_now)}"
                )
                if key not in rev_lists:
                    result = await anoncreds_registry.get_revocation_list(
                        self._profile,
                        rev_reg_id,
                        timestamp_from=reft_non_revoc_interval.get("from", 0),
                        timestamp_to=reft_non_revoc_interval.get("to", epoch_now),
                    )

                    rev_lists[key] = (
                        rev_reg_id,
                        credential_id,
                        result.revocation_list.serialize(),
                        result.revocation_list.timestamp,
                    )
                for stamp_me in requested_referents.values():
                    # often one cred satisfies many requested attrs/preds
                    if stamp_me["cred_id"] == credential_id:
                        stamp_me["timestamp"] = rev_lists[key][3]

        return rev_lists

    async def _get_revocation_states(
        self, revocation_registries: dict, credentials: dict, rev_lists: dict
    ):
        """Get revocation states to prove non-revoked."""
        revocation_states = {}
        for (
            rev_reg_id,
            credential_id,
            rev_list,
            timestamp,
        ) in rev_lists.values():
            if rev_reg_id not in revocation_states:
                revocation_states[rev_reg_id] = {}
            rev_reg_def = revocation_registries[rev_reg_id]
            revocation = AnonCredsRevocation(self._profile)
            tails_local_path = await revocation.get_or_fetch_local_tails_path(rev_reg_def)
            try:
                revocation_states[rev_reg_id][timestamp] = json.loads(
                    await self.holder.create_revocation_state(
                        credentials[credential_id]["cred_rev_id"],
                        rev_reg_def.serialize(),
                        rev_list,
                        tails_local_path,
                    )
                )
            except AnonCredsHolderError as e:
                LOGGER.error(
                    f"Failed to create revocation state: {e.error_code}, {e.message}"
                )
                raise e
        return revocation_states

    def _set_timestamps(self, requested_credentials: dict, requested_referents: dict):
        for referent, precis in requested_referents.items():
            if "timestamp" not in precis:
                continue
            if referent in requested_credentials["requested_attributes"]:
                requested_credentials["requested_attributes"][referent]["timestamp"] = (
                    precis["timestamp"]
                )
            if referent in requested_credentials["requested_predicates"]:
                requested_credentials["requested_predicates"][referent]["timestamp"] = (
                    precis["timestamp"]
                )

    async def return_presentation(
        self,
        pres_ex_record: Union[V10PresentationExchange, V20PresExRecord],
        requested_credentials: Optional[dict] = None,
    ) -> dict:
        """Return AnonCreds proof request as dict."""

        # If not anoncreds capable, try to use indy handler. This should be removed when
        # indy filter is completely retired
        if not isinstance(self._profile, AskarAnonCredsProfile):
            from ..indy.pres_exch_handler import IndyPresExchHandler

            handler = IndyPresExchHandler(self._profile)
            return await handler.return_presentation(
                pres_ex_record, requested_credentials
            )

        requested_credentials = requested_credentials or {}
        proof_request = self._extract_proof_request(pres_ex_record)
        non_revoc_intervals = extract_non_revocation_intervals_from_proof_request(
            proof_request
        )

        requested_referents = self._get_requested_referents(
            proof_request, requested_credentials, non_revoc_intervals
        )

        credentials = await self._get_credentials(requested_referents)
        self._remove_superfluous_timestamps(requested_credentials, credentials)

        schemas, cred_defs, revocation_registries = await self._get_ledger_objects(
            credentials
        )

        rev_lists = await self._get_revocation_lists(requested_referents, credentials)

        revocation_states = await self._get_revocation_states(
            revocation_registries, credentials, rev_lists
        )

        self._set_timestamps(requested_credentials, requested_referents)

        proof_json = await self.holder.create_presentation(
            proof_request,
            requested_credentials,
            schemas,
            cred_defs,
            revocation_states,
        )
        proof = json.loads(proof_json)
        return proof



"""mediate-request message used to request mediation from a mediator."""

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import MEDIATE_REQUEST, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.mediation_request_handler.MediationRequestHandler"
)


class MediationRequest(AgentMessage):
    """Represents a request for mediation."""

    class Meta:
        """MediationRequest metadata."""

        handler_class = HANDLER_CLASS
        message_type = MEDIATE_REQUEST
        schema_class = "MediationRequestSchema"

    def __init__(self, **kwargs):
        """Initialize mediation request object."""
        super(MediationRequest, self).__init__(**kwargs)


class MediationRequestSchema(AgentMessageSchema):
    """Mediation request schema class."""

    class Meta:
        """Mediation request schema metadata."""

        model_class = MediationRequest



"""Coordinate mediation problem report message handler."""

from .....messaging.base_handler import BaseHandler
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..messages.problem_report import CMProblemReport


class CMProblemReportHandler(BaseHandler):
    """Handler class for Coordinate Mediation Problem Report Message.

    Updates the ConnRecord Metadata state.
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Coordinate mediation problem report message handler.

        Args:
            context: Request context
            responder: Responder callback
        """
        self._logger.debug("CMProblemReportHandler called with context %s", context)
        assert isinstance(context.message, CMProblemReport)
        self._logger.error(
            f"Received coordinate-mediation problem report message: {context.message}"
        )



"""AnonCreds Schema OpenAPI validators."""

from typing import Any, Dict, List, Optional

from anoncreds import Schema
from marshmallow import EXCLUDE, fields
from marshmallow.validate import OneOf

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import ANONCREDS_DID_EXAMPLE, ANONCREDS_SCHEMA_ID_EXAMPLE


class AnonCredsSchema(BaseModel):
    """An AnonCreds Schema object."""

    class Meta:
        """AnonCredsSchema metadata."""

        schema_class = "AnonCredsSchemaSchema"

    def __init__(
        self, issuer_id: str, attr_names: List[str], name: str, version: str, **kwargs
    ):
        """Initialize an instance.

        Args:
            issuer_id: Issuer ID
            attr_names: Schema Attribute Name list
            name: Schema name
            version: Schema version
            kwargs: Addiiotnal keyword arguments
        """
        super().__init__(**kwargs)
        self.issuer_id = issuer_id
        self.attr_names = attr_names
        self.name = name
        self.version = version

    @classmethod
    def from_native(cls, schema: Schema) -> "AnonCredsSchema":
        """Convert from native object."""
        return cls.deserialize(schema.to_dict())

    def to_native(self) -> Schema:
        """Convert to native object."""
        return Schema.load(self.serialize())


class AnonCredsSchemaSchema(BaseModelSchema):
    """Marshmallow schema for anoncreds schema."""

    class Meta:
        """AnonCredsSchemaSchema metadata."""

        model_class = AnonCredsSchema
        unknown = EXCLUDE

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition or schema",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        data_key="issuerId",
        required=True,
    )
    attr_names = fields.List(
        fields.Str(
            metadata={
                "description": "Attribute name",
                "example": "score",
            }
        ),
        metadata={"description": "Schema attribute names"},
        data_key="attrNames",
        required=True,
    )
    name = fields.Str(
        metadata={"description": "Schema name", "example": "Example schema"},
        required=True,
    )
    version = fields.Str(
        metadata={"description": "Schema version", "example": "1.0"}, required=True
    )


class GetSchemaResult(BaseModel):
    """Result of resolving a schema."""

    class Meta:
        """GetSchemaResult metadata."""

        schema_class = "GetSchemaResultSchema"

    def __init__(
        self,
        schema: AnonCredsSchema,
        schema_id: str,
        resolution_metadata: Dict[str, Any],
        schema_metadata: Dict[str, Any],
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            schema: AnonCreds Schema
            schema_id: Schema ID
            resolution_metadata: Resolution Metdata
            schema_metadata: Schema Metadata
            kwargs: Additional keyword arguments
        """
        super().__init__(**kwargs)
        self.schema_value = schema
        self.schema_id = schema_id
        self.resolution_metadata = resolution_metadata
        self.schema_metadata = schema_metadata

    @property
    def schema(self) -> AnonCredsSchema:
        """Alias for schema_value.

        `schema` can't be used directly due to a limitation of marshmallow.
        """
        return self.schema_value


class GetSchemaResultSchema(BaseModelSchema):
    """Parameters and validators for schema create query."""

    class Meta:
        """GetSchemaResultSchema metadata."""

        model_class = GetSchemaResult
        unknown = EXCLUDE

    schema_value = fields.Nested(AnonCredsSchemaSchema(), data_key="schema")
    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        }
    )
    resolution_metadata = fields.Dict()
    schema_metadata = fields.Dict()


class SchemaState(BaseModel):
    """Model representing the state of a schema after beginning registration."""

    STATE_FINISHED = "finished"
    STATE_FAILED = "failed"
    STATE_ACTION = "action"
    STATE_WAIT = "wait"

    class Meta:
        """SchemaState metadata."""

        schema_class = "SchemaStateSchema"

    def __init__(self, state: str, schema_id: str, schema: AnonCredsSchema, **kwargs):
        """Initialize a new SchemaState."""
        super().__init__(**kwargs)
        self.state = state
        self.schema_id = schema_id
        self.schema_value = schema

    @property
    def schema(self) -> AnonCredsSchema:
        """Alias to schema_value.

        `schema` can't be used directly due to limitations of marshmallow.
        """
        return self.schema_value


class SchemaStateSchema(BaseModelSchema):
    """Parameters and validators for schema state."""

    class Meta:
        """SchemaStateSchema metadata."""

        model_class = SchemaState

    state = fields.Str(
        validate=OneOf(
            [
                SchemaState.STATE_FINISHED,
                SchemaState.STATE_FAILED,
                SchemaState.STATE_ACTION,
                SchemaState.STATE_WAIT,
            ]
        )
    )
    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        }
    )
    schema_value = fields.Nested(AnonCredsSchemaSchema(), data_key="schema")


class SchemaResult(BaseModel):
    """Result of registering a schema."""

    class Meta:
        """SchemaResult metadata."""

        schema_class = "SchemaResultSchema"

    def __init__(
        self,
        job_id: Optional[str],
        schema_state: SchemaState,
        registration_metadata: Optional[dict] = None,
        schema_metadata: Optional[dict] = None,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            job_id: Job ID
            schema_state: Schema state
            registration_metadata: Registration Metdata
            schema_metadata: Schema Metadata
            kwargs: Additional keyword arguments
        """
        super().__init__(**kwargs)
        self.job_id = job_id
        self.schema_state = schema_state
        self.registration_metadata = registration_metadata or {}
        self.schema_metadata = schema_metadata or {}


class SchemaResultSchema(BaseModelSchema):
    """Parameters and validators for schema state."""

    class Meta:
        """SchemaResultSchema metadata."""

        model_class = SchemaResult

    job_id = fields.Str()
    schema_state = fields.Nested(SchemaStateSchema())
    registration_metadata = fields.Dict()
    # For indy, schema_metadata will contain the seqNo
    schema_metadata = fields.Dict()



from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import CRED_20_PROBLEM_REPORT, PROTOCOL_PACKAGE
from .. import cred_problem_report as test_module
from ..cred_problem_report import (
    ProblemReportReason,
    V20CredProblemReport,
    V20CredProblemReportSchema,
    ValidationError,
)


class TestCredProblemReport(TestCase):
    """Problem report tests."""

    def test_init_type(self):
        """Test initializer."""

        prob = V20CredProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        assert prob._type == DIDCommPrefix.qualify_current(CRED_20_PROBLEM_REPORT)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.cred_problem_report.V20CredProblemReportSchema.load"
    )
    def test_deserialize(self, mock_load):
        """Test deserialization."""

        obj = V20CredProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )

        prob = V20CredProblemReport.deserialize(obj)
        mock_load.assert_called_once_with(obj)

        assert prob is mock_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.cred_problem_report.V20CredProblemReportSchema.dump"
    )
    def test_serialize(self, mock_dump):
        """Test serialization."""

        obj = V20CredProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )

        ser = obj.serialize()
        mock_dump.assert_called_once_with(obj)

        assert ser is mock_dump.return_value

    def test_make_model(self):
        """Test making model."""

        prob = V20CredProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        data = prob.serialize()
        model_instance = V20CredProblemReport.deserialize(data)
        assert isinstance(model_instance, V20CredProblemReport)

        prob = V20CredProblemReport()
        data = prob.serialize()
        with pytest.raises(BaseModelError):
            V20CredProblemReport.deserialize(data)

    def test_validate_and_logger(self):
        """Capture ValidationError and Logs."""
        data = V20CredProblemReport(
            description={
                "en": "Insufficient credit",
                "code": "invalid_code",
            },
        ).serialize()
        with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
            V20CredProblemReportSchema().validate_fields(data)
        assert mock_logger.warning.call_count == 1

    def test_validate_x(self):
        """Exercise validation requirements."""
        schema = V20CredProblemReportSchema()
        with pytest.raises(ValidationError):
            schema.validate_fields({})



TEST_LD_DOCUMENT_BAD_SIGNED_BBS = {
    "@context": [
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
        "https://w3id.org/security/bbs/v1",
    ],
    "name": "John Doe",
    "homepage": "https://domain.com/profile",
    "proof": {
        "type": "BbsBlsSignature2020",
        "created": "2020-04-18T02:41:11Z",
        "verificationMethod": "did:example:489398593#test",
        "proofPurpose": "assertionMethod",
        "signature": "cBJG+AYOEuLd3MwsOC1xyPRcmspso687W24zSvduShKbL0xCJk7f0rrEpKkBpYPHThU3f3gP0d4amCQ9QLkGT5Z9mw2r2MakfdliAhnVt/CvV1IEEyouFAqzCiEyfxhU1wAAAAAAAAAAAAAAAAAAAABT3s/8eZfUJrJNLFxMiMnIhYWQvkqA80eRpwTZKWjRUwAAAAAAAAAAAAAAAAAAAAADp+HUtElIe+t1Pb0tiqrodTEb+pU4i9tDihJBY0xODA==",
    },
}



"""Handle connection information interface with non-secrets storage."""

import json
from enum import Enum
from typing import Any, Optional, Union

from marshmallow import fields, validate

from ...core.profile import ProfileSession
from ...messaging.models.base_record import BaseRecord, BaseRecordSchema
from ...messaging.valid import (
    GENERIC_DID_EXAMPLE,
    GENERIC_DID_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
    UUID4_EXAMPLE,
)
from ...protocols.didexchange.v1_0.message_types import ARIES_PROTOCOL as DIDEX_1_1
from ...protocols.didexchange.v1_0.message_types import DIDEX_1_0
from ...protocols.didexchange.v1_0.messages.request import DIDXRequest
from ...protocols.out_of_band.v1_0.messages.invitation import (
    InvitationMessage as OOBInvitation,
)
from ...storage.base import BaseStorage
from ...storage.error import StorageNotFoundError
from ...storage.record import StorageRecord


class ConnRecord(BaseRecord):
    """Represents a single pairwise connection."""

    class Meta:
        """ConnRecord metadata."""

        schema_class = "MaybeStoredConnRecordSchema"

    SUPPORTED_PROTOCOLS = (DIDEX_1_0, DIDEX_1_1)

    class Role(Enum):
        """RFC 160 (inviter, invitee) = RFC 23 (responder, requester)."""

        REQUESTER = ("invitee", "requester")  # == RFC 23 initiator, RFC 434 receiver
        RESPONDER = ("inviter", "responder")  # == RFC 160 initiator(!), RFC 434 sender

        @property
        def rfc160(self):
            """Return RFC 160 (connection protocol) nomenclature."""
            return self.value[0]

        @property
        def rfc23(self):
            """Return RFC 23 (DID exchange protocol) nomenclature."""
            return self.value[1]

        @classmethod
        def get(cls, label: Union[str, "ConnRecord.Role"]):
            """Get role enum for label."""
            if isinstance(label, str):
                for role in ConnRecord.Role:
                    if label in role.value:
                        return role
            elif isinstance(label, ConnRecord.Role):
                return label
            return None

        def flip(self):
            """Return opposite interlocutor role: theirs for ours, ours for theirs."""
            return (
                ConnRecord.Role.REQUESTER
                if self is ConnRecord.Role.RESPONDER
                else ConnRecord.Role.RESPONDER
            )

        def __eq__(self, other: Union[str, "ConnRecord.Role"]) -> bool:
            """Comparison between roles."""
            return self is ConnRecord.Role.get(other)

    class State(Enum):
        """Collator for equivalent states between RFC 160 and RFC 23.

        On the connection record, the state has to serve for both RFCs.
        Hence, internally, RFC23 requester/responder states collate to
        their RFC160 condensed equivalent.
        """

        INIT = ("init", "start")
        INVITATION = ("invitation", "invitation")
        REQUEST = ("request", "request")
        RESPONSE = ("response", "response")
        COMPLETED = ("active", "completed")
        ABANDONED = ("error", "abandoned")

        @property
        def rfc160(self):
            """Return RFC 160 (connection protocol) nomenclature."""
            return self.value[0]

        @property
        def rfc23(self):
            """Return RFC 23 (DID exchange protocol) nomenclature to record logic."""
            return self.value[1]

        def rfc23strict(self, their_role: "ConnRecord.Role"):
            """Return RFC 23 (DID exchange protocol) nomenclature to role as per RFC."""

            if not their_role or self in (
                ConnRecord.State.INIT,
                ConnRecord.State.COMPLETED,
                ConnRecord.State.ABANDONED,
            ):
                return self.value[1]

            if self is ConnRecord.State.REQUEST:
                return self.value[1] + (
                    "-sent"
                    if ConnRecord.Role.get(their_role) is ConnRecord.Role.RESPONDER
                    else "-received"
                )
            else:
                return self.value[1] + (
                    "-received"
                    if ConnRecord.Role.get(their_role) is ConnRecord.Role.RESPONDER
                    else "-sent"
                )

        @classmethod
        def get(cls, label: Union[str, "ConnRecord.State"]):
            """Get state enum for label."""
            if isinstance(label, str):
                for state in ConnRecord.State:
                    if label in state.value:
                        return state
            elif isinstance(label, ConnRecord.State):
                return label
            return None

        def __eq__(self, other: Union[str, "ConnRecord.State"]) -> bool:
            """Comparison between states."""
            return self is ConnRecord.State.get(other)

    RECORD_ID_NAME = "connection_id"
    RECORD_TOPIC = "connections"
    LOG_STATE_FLAG = "debug.connections"
    TAG_NAMES = {
        "my_did",
        "their_did",
        "request_id",
        "invitation_key",
        "their_public_did",
        "invitation_msg_id",
        "state",
        "their_role",
    }

    RECORD_TYPE = "connection"
    RECORD_TYPE_INVITATION = "connection_invitation"
    RECORD_TYPE_REQUEST = "connection_request"
    RECORD_TYPE_METADATA = "connection_metadata"

    INVITATION_MODE_ONCE = "once"
    INVITATION_MODE_MULTI = "multi"
    INVITATION_MODE_STATIC = "static"

    ACCEPT_MANUAL = "manual"
    ACCEPT_AUTO = "auto"

    def __init__(
        self,
        *,
        connection_id: Optional[str] = None,
        my_did: Optional[str] = None,
        their_did: Optional[str] = None,
        their_label: Optional[str] = None,
        their_role: Union[str, "ConnRecord.Role", None] = None,
        invitation_key: Optional[str] = None,
        invitation_msg_id: Optional[str] = None,
        request_id: Optional[str] = None,
        state: Union[str, "ConnRecord.State", None] = None,
        inbound_connection_id: Optional[str] = None,
        error_msg: Optional[str] = None,
        accept: Optional[str] = None,
        invitation_mode: Optional[str] = None,
        alias: Optional[str] = None,
        their_public_did: Optional[str] = None,
        connection_protocol: Optional[str] = None,
        # from state: formalism for base_record.from_storage()
        rfc23_state: Optional[str] = None,
        # for backward compat with old records
        routing_state: Optional[str] = None,
        # for backward compatibility with old ConnectionRecord
        initiator: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a new ConnRecord."""
        super().__init__(
            connection_id,
            state=(ConnRecord.State.get(state) or ConnRecord.State.INIT).rfc160,
            **kwargs,
        )
        self.my_did = my_did
        self.their_did = their_did
        self.their_label = their_label
        self.their_role = (
            ConnRecord.Role.get(their_role).rfc160
            if isinstance(their_role, str)
            else None
            if their_role is None
            else their_role.rfc160
        )
        self.invitation_key = invitation_key
        self.invitation_msg_id = invitation_msg_id
        self.request_id = request_id
        self.error_msg = error_msg
        self.inbound_connection_id = inbound_connection_id
        self.accept = accept or self.ACCEPT_MANUAL
        self.invitation_mode = invitation_mode or self.INVITATION_MODE_ONCE
        self.alias = alias
        self.their_public_did = their_public_did
        self.connection_protocol = (
            connection_protocol
            if connection_protocol in self.SUPPORTED_PROTOCOLS
            else None
        )

    @property
    def connection_id(self) -> str:
        """Accessor for the ID associated with this connection."""
        return self._id

    @property
    def rfc23_state(self) -> str:
        """RFC23 state per RFC text, potentially particular to role."""
        return ConnRecord.State.get(self.state).rfc23strict(self.their_role)

    @property
    def record_value(self) -> dict:
        """Accessor to for the JSON record value properties for this connection."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "their_role",
                "inbound_connection_id",
                "accept",
                "invitation_mode",
                "invitation_msg_id",
                "alias",
                "error_msg",
                "their_label",
                "their_public_did",
                "state",
                "connection_protocol",
            )
        }

    @classmethod
    async def retrieve_by_did(
        cls,
        session: ProfileSession,
        their_did: Optional[str] = None,
        my_did: Optional[str] = None,
        their_role: Optional[str] = None,
    ) -> "ConnRecord":
        """Retrieve a connection record by target DID.

        Args:
            session: The active profile session
            their_did: The target DID to filter by
            my_did: One of our DIDs to filter by
            my_role: Filter connections by their role
            their_role: Filter connections by their role
        """
        tag_filter = {}
        if their_did:
            tag_filter["their_did"] = their_did
        if my_did:
            tag_filter["my_did"] = my_did

        post_filter = {}
        if their_role:
            post_filter["their_role"] = cls.Role.get(their_role).rfc160

        return await cls.retrieve_by_tag_filter(session, tag_filter, post_filter)

    @classmethod
    async def retrieve_by_did_peer_4(
        cls,
        session: ProfileSession,
        their_did_long: Optional[str] = None,
        their_did_short: Optional[str] = None,
        my_did: Optional[str] = None,
        their_role: Optional[str] = None,
    ) -> "ConnRecord":
        """Retrieve a connection record by target DID.

        Args:
            session: The active profile session
            their_did_long: The target DID to filter by, in long form
            their_did_short: The target DID to filter by, in short form
            my_did: One of our DIDs to filter by
            my_role: Filter connections by their role
            their_role: Filter connections by their role
        """
        tag_filter = {}
        if their_did_long and their_did_short:
            tag_filter["$or"] = [
                {"their_did": their_did_long},
                {"their_did": their_did_short},
            ]
        elif their_did_short:
            tag_filter["their_did"] = their_did_short
        elif their_did_long:
            tag_filter["their_did"] = their_did_long
        if my_did:
            tag_filter["my_did"] = my_did

        post_filter = {}
        if their_role:
            post_filter["their_role"] = cls.Role.get(their_role).rfc160

        return await cls.retrieve_by_tag_filter(session, tag_filter, post_filter)

    @classmethod
    async def retrieve_by_invitation_key(
        cls,
        session: ProfileSession,
        invitation_key: str,
        their_role: Optional[str] = None,
    ) -> "ConnRecord":
        """Retrieve a connection record by invitation key.

        Args:
            session: The active profile session
            invitation_key: The key on the originating invitation
            their_role: Filter by their role
        """
        tag_filter = {
            "invitation_key": invitation_key,
            "state": cls.State.INVITATION.rfc160,
        }
        post_filter = {"state": cls.State.INVITATION.rfc160}

        if their_role:
            post_filter["their_role"] = cls.Role.get(their_role).rfc160
            tag_filter["their_role"] = cls.Role.get(their_role).rfc160

        return await cls.retrieve_by_tag_filter(session, tag_filter, post_filter)

    @classmethod
    async def retrieve_by_invitation_msg_id(
        cls,
        session: ProfileSession,
        invitation_msg_id: str,
        their_role: Optional[str] = None,
    ) -> Optional["ConnRecord"]:
        """Retrieve a connection record by invitation_msg_id.

        Args:
            session: The active profile session
            invitation_msg_id: Invitation message identifier
            their_role: Filter by their role
        """
        tag_filter = {"invitation_msg_id": invitation_msg_id}
        post_filter = {
            "state": cls.State.INVITATION.rfc160,
        }
        if their_role:
            post_filter["their_role"] = cls.Role.get(their_role).rfc160
        try:
            return await cls.retrieve_by_tag_filter(session, tag_filter, post_filter)
        except StorageNotFoundError:
            return None

    @classmethod
    async def find_existing_connection(
        cls, session: ProfileSession, their_public_did: str
    ) -> Optional["ConnRecord"]:
        """Retrieve existing active connection records (public did or did:peer).

        Args:
            session: The active profile session
            their_public_did: Inviter public DID (or did:peer)
        """
        tag_filter = {"their_public_did": their_public_did}
        conn_records = await cls.query(
            session,
            tag_filter=tag_filter,
        )
        for conn_record in conn_records:
            if conn_record.state == ConnRecord.State.COMPLETED:
                return conn_record
        return None

    @classmethod
    async def retrieve_by_request_id(
        cls, session: ProfileSession, request_id: str, their_role: Optional[str] = None
    ) -> "ConnRecord":
        """Retrieve a connection record from our previous request ID.

        Args:
            session: The active profile session
            request_id: The ID of the originating connection request
            their_role: Filter by their role
        """
        tag_filter = {"request_id": request_id}
        if their_role:
            tag_filter["their_role"] = their_role
        return await cls.retrieve_by_tag_filter(session, tag_filter)

    @classmethod
    async def retrieve_by_alias(cls, session: ProfileSession, alias: str) -> "ConnRecord":
        """Retrieve a connection record from an alias.

        Args:
            session: The active profile session
            alias: The alias of the connection
        """
        post_filter = {"alias": alias}
        return await cls.query(session, post_filter_positive=post_filter)

    async def attach_invitation(
        self,
        session: ProfileSession,
        invitation: OOBInvitation,
    ):
        """Persist the related connection invitation to storage.

        Args:
            session: The active profile session
            invitation: The invitation to relate to this connection record
        """
        assert self.connection_id
        record = StorageRecord(
            self.RECORD_TYPE_INVITATION,  # conn- or oob-invitation, to retrieve easily
            invitation.to_json(),
            {"connection_id": self.connection_id},
        )
        storage = session.inject(BaseStorage)
        await storage.add_record(record)

    async def retrieve_invitation(self, session: ProfileSession) -> OOBInvitation:
        """Retrieve the related connection invitation.

        Args:
            session: The active profile session
        """
        assert self.connection_id
        storage = session.inject(BaseStorage)
        result = await storage.find_record(
            self.RECORD_TYPE_INVITATION,
            {"connection_id": self.connection_id},
        )
        ser = json.loads(result.value)
        return OOBInvitation.deserialize(ser)

    async def attach_request(
        self,
        session: ProfileSession,
        request: DIDXRequest,
    ):
        """Persist the related connection request to storage.

        Args:
            session: The active profile session
            request: The request to relate to this connection record
        """
        assert self.connection_id
        record = StorageRecord(
            self.RECORD_TYPE_REQUEST,  # conn- or didx-request, to retrieve easily
            request.to_json(),
            {"connection_id": self.connection_id},
        )
        storage: BaseStorage = session.inject(BaseStorage)
        await storage.add_record(record)

    async def retrieve_request(
        self,
        session: ProfileSession,
    ) -> DIDXRequest:
        """Retrieve the related connection invitation.

        Args:
            session: The active profile session
        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        result = await storage.find_record(
            self.RECORD_TYPE_REQUEST, {"connection_id": self.connection_id}
        )
        ser = json.loads(result.value)
        return DIDXRequest.deserialize(ser)

    @property
    def is_ready(self) -> str:
        """Accessor for connection readiness."""
        return ConnRecord.State.get(self.state) in (
            ConnRecord.State.COMPLETED,
            ConnRecord.State.RESPONSE,
        )

    @property
    def is_multiuse_invitation(self) -> bool:
        """Accessor for multi use invitation mode."""
        return self.invitation_mode == self.INVITATION_MODE_MULTI

    async def post_save(self, session: ProfileSession, *args, **kwargs):
        """Perform post-save actions.

        Args:
            session: The active profile session
            args: Additional positional arguments
            kwargs: Additional keyword arguments
        """
        await super().post_save(session, *args, **kwargs)

        # clear cache key set by connection manager
        cache_key = f"connection_target::{self.connection_id}"
        await self.clear_cached_key(session, cache_key)

    async def delete_record(self, session: ProfileSession):
        """Perform connection record deletion actions.

        Args:
            session (ProfileSession): session

        """
        await super().delete_record(session)

        storage = session.inject(BaseStorage)
        # Delete metadata
        if self.connection_id:
            await storage.delete_all_records(
                self.RECORD_TYPE_METADATA,
                {"connection_id": self.connection_id},
            )

        # Delete attached messages
        await storage.delete_all_records(
            self.RECORD_TYPE_REQUEST,
            {"connection_id": self.connection_id},
        )
        await storage.delete_all_records(
            self.RECORD_TYPE_INVITATION,
            {"connection_id": self.connection_id},
        )

    async def abandon(self, session: ProfileSession, *, reason: Optional[str] = None):
        """Set state to abandoned."""
        reason = reason or "Connection abandoned"
        self.state = ConnRecord.State.ABANDONED.rfc160
        self.error_msg = reason
        await self.save(session, reason=reason)

    async def metadata_get(
        self, session: ProfileSession, key: str, default: Optional[Any] = None
    ) -> Any:
        """Retrieve arbitrary metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key identifying metadata
            default (Any): default value to get; type should be a JSON
                compatible value.

        Returns:
            Any: metadata stored by key

        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            return json.loads(record.value)
        except StorageNotFoundError:
            return default

    async def metadata_set(self, session: ProfileSession, key: str, value: Any):
        """Set arbitrary metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key identifying metadata
            value (Any): value to set
        """
        assert self.connection_id
        value = json.dumps(value)
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.update_record(record, value, record.tags)
        except StorageNotFoundError:
            record = StorageRecord(
                self.RECORD_TYPE_METADATA,
                value,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.add_record(record)

    async def metadata_delete(self, session: ProfileSession, key: str):
        """Delete custom metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key of metadata to delete
        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.delete_record(record)
        except StorageNotFoundError as err:
            raise KeyError(f"{key} not found in connection metadata") from err

    async def metadata_get_all(self, session: ProfileSession) -> dict:
        """Return all custom metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage

        Returns:
            dict: dictionary representation of all metadata values

        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        records = await storage.find_all_records(
            self.RECORD_TYPE_METADATA,
            {"connection_id": self.connection_id},
        )
        return {record.tags["key"]: json.loads(record.value) for record in records}

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class MaybeStoredConnRecordSchema(BaseRecordSchema):
    """Schema to allow serialization/deserialization of connection records."""

    class Meta:
        """MaybeStoredConnRecordSchema metadata."""

        model_class = ConnRecord

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    my_did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={
            "description": "Our DID for connection",
            "example": GENERIC_DID_EXAMPLE,
        },
    )
    their_did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={
            "description": "Their DID for connection",
            "example": GENERIC_DID_EXAMPLE,
        },
    )
    their_label = fields.Str(
        required=False,
        metadata={"description": "Their label for connection", "example": "Bob"},
    )
    their_role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [label for role in ConnRecord.Role for label in role.value]
        ),
        metadata={
            "description": "Their role in the connection protocol",
            "example": ConnRecord.Role.REQUESTER.rfc23,
        },
    )
    connection_protocol = fields.Str(
        required=False,
        validate=validate.OneOf(ConnRecord.SUPPORTED_PROTOCOLS),
        metadata={
            "description": "Connection protocol used",
            "example": "didexchange/1.1",
        },
    )
    rfc23_state = fields.Str(
        dump_only=True,
        metadata={"description": "State per RFC 23", "example": "invitation-sent"},
    )
    inbound_connection_id = fields.Str(
        required=False,
        metadata={
            "description": "Inbound routing connection id to use",
            "example": UUID4_EXAMPLE,
        },
    )
    invitation_key = fields.Str(
        required=False,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Public key for connection",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    invitation_msg_id = fields.Str(
        required=False,
        metadata={
            "description": "ID of out-of-band invitation message",
            "example": UUID4_EXAMPLE,
        },
    )
    request_id = fields.Str(
        required=False,
        metadata={
            "description": "Connection request identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    accept = fields.Str(
        required=False,
        validate=validate.OneOf(
            ConnRecord.get_attributes_by_prefix("ACCEPT_", walk_mro=False)
        ),
        metadata={
            "description": "Connection acceptance: manual or auto",
            "example": ConnRecord.ACCEPT_AUTO,
        },
    )
    error_msg = fields.Str(
        required=False,
        metadata={
            "description": "Error message",
            "example": "No DIDDoc provided; cannot connect to public DID",
        },
    )
    invitation_mode = fields.Str(
        required=False,
        validate=validate.OneOf(
            ConnRecord.get_attributes_by_prefix("INVITATION_MODE_", walk_mro=False)
        ),
        metadata={
            "description": "Invitation mode",
            "example": ConnRecord.INVITATION_MODE_ONCE,
        },
    )
    alias = fields.Str(
        required=False,
        metadata={
            "description": "Optional alias to apply to connection for later use",
            "example": "Bob, providing quotes",
        },
    )
    their_public_did = fields.Str(
        required=False,
        metadata={
            "description": "Other agent's public DID for connection",
            "example": "2cpBmR3FqGKWi5EyUbpRY8",
        },
    )


class ConnRecordSchema(MaybeStoredConnRecordSchema):
    """Schema representing stored ConnRecords."""

    class Meta:
        """ConnRecordSchema metadata."""

        model_class = ConnRecord

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )



"""Aries#0036 v1.0 credential exchange information with non-secrets storage."""

import logging
from typing import Any, Mapping, Optional, Union

from marshmallow import fields, validate

from .....core.profile import ProfileSession
from .....indy.models.cred import IndyCredential, IndyCredentialSchema
from .....indy.models.cred_abstract import IndyCredAbstract, IndyCredAbstractSchema
from .....indy.models.cred_precis import IndyCredInfo, IndyCredInfoSchema
from .....indy.models.cred_request import IndyCredRequest, IndyCredRequestSchema
from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    UUID4_EXAMPLE,
)
from .....storage.base import StorageError
from ..messages.credential_exchange_webhook import V10CredentialExchangeWebhook
from ..messages.credential_offer import CredentialOffer, CredentialOfferSchema
from ..messages.credential_proposal import CredentialProposal, CredentialProposalSchema
from . import UNENCRYPTED_TAGS

LOGGER = logging.getLogger(__name__)


class V10CredentialExchange(BaseExchangeRecord):
    """Represents an Aries#0036 credential exchange."""

    class Meta:
        """CredentialExchange metadata."""

        schema_class = "V10CredentialExchangeSchema"

    RECORD_TYPE = "credential_exchange_v10"
    RECORD_ID_NAME = "credential_exchange_id"
    RECORD_TOPIC = "issue_credential"
    TAG_NAMES = {"~thread_id"} if UNENCRYPTED_TAGS else {"thread_id"}

    INITIATOR_SELF = "self"
    INITIATOR_EXTERNAL = "external"
    ROLE_ISSUER = "issuer"
    ROLE_HOLDER = "holder"

    STATE_PROPOSAL_SENT = "proposal_sent"
    STATE_PROPOSAL_RECEIVED = "proposal_received"
    STATE_OFFER_SENT = "offer_sent"
    STATE_OFFER_RECEIVED = "offer_received"
    STATE_REQUEST_SENT = "request_sent"
    STATE_REQUEST_RECEIVED = "request_received"
    STATE_ISSUED = "credential_issued"
    STATE_CREDENTIAL_RECEIVED = "credential_received"
    STATE_ACKED = "credential_acked"
    STATE_CREDENTIAL_REVOKED = "credential_revoked"
    STATE_ABANDONED = "abandoned"

    def __init__(
        self,
        *,
        credential_exchange_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        parent_thread_id: Optional[str] = None,
        initiator: Optional[str] = None,
        role: Optional[str] = None,
        state: Optional[str] = None,
        credential_definition_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        credential_proposal_dict: Union[
            Mapping, CredentialProposal
        ] = None,  # aries message: ..._dict for historic compat on all aries msgs
        credential_offer_dict: Union[Mapping, CredentialOffer] = None,  # aries message
        credential_offer: Union[Mapping, IndyCredAbstract] = None,  # indy artifact
        credential_request: Union[Mapping, IndyCredRequest] = None,  # indy artifact
        credential_request_metadata: Optional[Mapping] = None,
        credential_id: Optional[str] = None,
        raw_credential: Union[Mapping, IndyCredential] = None,  # indy cred as received
        credential: Union[Mapping, IndyCredInfo] = None,  # indy cred as stored
        revoc_reg_id: Optional[str] = None,
        revocation_id: Optional[str] = None,
        auto_offer: bool = False,
        auto_issue: bool = False,
        auto_remove: bool = True,
        error_msg: Optional[str] = None,
        trace: bool = False,  # backward-compat: BaseRecord.from_storage()
        **kwargs,
    ):
        """Initialize a new V10CredentialExchange."""
        super().__init__(credential_exchange_id, state, trace=trace, **kwargs)
        self._id = credential_exchange_id
        self.connection_id = connection_id
        self.thread_id = thread_id
        self.parent_thread_id = parent_thread_id
        self.initiator = initiator
        self.role = role
        self.state = state
        self.credential_definition_id = credential_definition_id
        self.schema_id = schema_id
        self._credential_proposal_dict = CredentialProposal.serde(
            credential_proposal_dict
        )
        self._credential_offer_dict = CredentialOffer.serde(credential_offer_dict)
        self._credential_offer = IndyCredAbstract.serde(credential_offer)
        self._credential_request = IndyCredRequest.serde(credential_request)
        self.credential_request_metadata = credential_request_metadata
        self.credential_id = credential_id
        self._raw_credential = IndyCredential.serde(raw_credential)
        self._credential = IndyCredInfo.serde(credential)
        self.revoc_reg_id = revoc_reg_id
        self.revocation_id = revocation_id
        self.auto_offer = auto_offer
        self.auto_issue = auto_issue
        self.auto_remove = auto_remove
        self.error_msg = error_msg

    @property
    def credential_exchange_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def credential_proposal_dict(self) -> CredentialProposal:
        """Accessor; get deserialized view."""
        return (
            None
            if self._credential_proposal_dict is None
            else self._credential_proposal_dict.de
        )

    @credential_proposal_dict.setter
    def credential_proposal_dict(self, value):
        """Setter; store de/serialized views."""
        self._credential_proposal_dict = CredentialProposal.serde(value)

    @property
    def credential_offer_dict(self) -> CredentialOffer:
        """Accessor; get deserialized view."""
        return (
            None
            if self._credential_offer_dict is None
            else self._credential_offer_dict.de
        )

    @credential_offer_dict.setter
    def credential_offer_dict(self, value):
        """Setter; store de/serialized views."""
        self._credential_offer_dict = CredentialOffer.serde(value)

    @property
    def credential_offer(self) -> IndyCredAbstract:
        """Accessor; get deserialized view."""
        return None if self._credential_offer is None else self._credential_offer.de

    @credential_offer.setter
    def credential_offer(self, value):
        """Setter; store de/serialized views."""
        self._credential_offer = IndyCredAbstract.serde(value)

    @property
    def credential_request(self) -> IndyCredRequest:
        """Accessor; get deserialized view."""
        return None if self._credential_request is None else self._credential_request.de

    @credential_request.setter
    def credential_request(self, value):
        """Setter; store de/serialized views."""
        self._credential_request = IndyCredRequest.serde(value)

    @property
    def raw_credential(self) -> IndyCredential:
        """Accessor; get deserialized view."""
        return None if self._raw_credential is None else self._raw_credential.de

    @raw_credential.setter
    def raw_credential(self, value):
        """Setter; store de/serialized views."""
        self._raw_credential = IndyCredential.serde(value)

    @property
    def credential(self) -> IndyCredInfo:
        """Accessor; get deserialized view."""
        return None if self._credential is None else self._credential.de

    @credential.setter
    def credential(self, value):
        """Setter; store de/serialized views."""
        self._credential = IndyCredInfo.serde(value)

    async def save_error_state(
        self,
        session: ProfileSession,
        *,
        state: Optional[str] = None,
        reason: Optional[str] = None,
        log_params: Mapping[str, Any] = None,
        log_override: bool = False,
    ):
        """Save record error state if need be; log and swallow any storage error.

        Args:
            session: The profile session to use
            state: The state to set
            reason: A reason to add to the log
            log_params: Additional parameters to log
            log_override: Override configured logging regimen, print to stderr instead
        """

        if self._last_state == state:  # already done
            return

        self.state = state or V10CredentialExchange.STATE_ABANDONED
        if reason:
            self.error_msg = reason

        try:
            await self.save(
                session,
                reason=reason,
                log_params=log_params,
                log_override=log_override,
            )
        except StorageError:
            LOGGER.exception("Error saving credential exchange error state")

    # Override
    async def emit_event(self, session: ProfileSession, payload: Optional[Any] = None):
        """Emit an event.

        Args:
            session: The profile session to use
            payload: The event payload
        """

        if not self.RECORD_TOPIC:
            return

        if self.state:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}::{self.state}"
        else:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}"

        if session.profile.settings.get("debug.webhooks"):
            if not payload:
                payload = self.serialize()
        else:
            payload = V10CredentialExchangeWebhook(**self.__dict__)
            payload = payload.__dict__

        await session.profile.notify(topic, payload)

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this invitation."""
        return {
            **{
                prop: getattr(self, prop)
                for prop in (
                    "connection_id",
                    "credential_request_metadata",
                    "error_msg",
                    "auto_offer",
                    "auto_issue",
                    "auto_remove",
                    "parent_thread_id",
                    "initiator",
                    "credential_definition_id",
                    "schema_id",
                    "credential_id",
                    "revoc_reg_id",
                    "revocation_id",
                    "role",
                    "state",
                    "trace",
                )
            },
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in (
                    "credential_proposal_dict",
                    "credential_offer_dict",
                    "credential_offer",
                    "credential_request",
                    "raw_credential",
                    "credential",
                )
                if getattr(self, prop) is not None
            },
        }

    @classmethod
    async def retrieve_by_connection_and_thread(
        cls,
        session: ProfileSession,
        connection_id: Optional[str],
        thread_id: str,
        role: Optional[str] = None,
        *,
        for_update=False,
    ) -> "V10CredentialExchange":
        """Retrieve a credential exchange record by connection and thread ID."""
        cache_key = f"credential_exchange_ctidx::{connection_id}::{thread_id}::{role}"
        record_id = await cls.get_cached_key(session, cache_key)
        if record_id:
            record = await cls.retrieve_by_id(session, record_id, for_update=for_update)
        else:
            post_filter = {}
            if role:
                post_filter["role"] = role
            if connection_id:
                post_filter["connection_id"] = connection_id
            record = await cls.retrieve_by_tag_filter(
                session,
                {"thread_id": thread_id},
                post_filter,
                for_update=for_update,
            )
            await cls.set_cached_key(session, cache_key, record.credential_exchange_id)
        return record

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V10CredentialExchangeSchema(BaseExchangeSchema):
    """Schema to allow serialization/deserialization of credential exchange records."""

    class Meta:
        """V10CredentialExchangeSchema metadata."""

        model_class = V10CredentialExchange

    credential_exchange_id = fields.Str(
        required=False,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    parent_thread_id = fields.Str(
        required=False,
        metadata={
            "description": "Parent thread identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    initiator = fields.Str(
        required=False,
        validate=validate.OneOf(["self", "external"]),
        metadata={
            "description": "Issue-credential exchange initiator: self or external",
            "example": V10CredentialExchange.INITIATOR_SELF,
        },
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(["holder", "issuer"]),
        metadata={
            "description": "Issue-credential exchange role: holder or issuer",
            "example": V10CredentialExchange.ROLE_ISSUER,
        },
    )
    state = fields.Str(
        required=False,
        metadata={
            "description": "Issue-credential exchange state",
            "example": V10CredentialExchange.STATE_ACKED,
        },
    )
    credential_definition_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    credential_proposal_dict = fields.Nested(
        CredentialProposalSchema(),
        required=False,
        metadata={"description": "Credential proposal message"},
    )
    credential_offer_dict = fields.Nested(
        CredentialOfferSchema(),
        required=False,
        metadata={"description": "Credential offer message"},
    )
    credential_offer = fields.Nested(
        IndyCredAbstractSchema(),
        required=False,
        metadata={"description": "(Indy) credential offer"},
    )
    credential_request = fields.Nested(
        IndyCredRequestSchema(),
        required=False,
        metadata={"description": "(Indy) credential request"},
    )
    credential_request_metadata = fields.Dict(
        required=False, metadata={"description": "(Indy) credential request metadata"}
    )
    credential_id = fields.Str(
        required=False,
        metadata={"description": "Credential identifier", "example": UUID4_EXAMPLE},
    )
    raw_credential = fields.Nested(
        IndyCredentialSchema(),
        required=False,
        metadata={
            "description": "Credential as received, prior to storage in holder wallet"
        },
    )
    credential = fields.Nested(
        IndyCredInfoSchema(),
        required=False,
        metadata={"description": "Credential as stored"},
    )
    auto_offer = fields.Bool(
        required=False,
        metadata={
            "description": "Holder choice to accept offer in this credential exchange",
            "example": False,
        },
    )
    auto_issue = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Issuer choice to issue to request in this credential exchange"
            ),
            "example": False,
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=True,
        metadata={
            "description": (
                "Issuer choice to remove this credential exchange record when complete"
            ),
            "example": False,
        },
    )
    error_msg = fields.Str(
        required=False,
        metadata={
            "description": "Error message",
            "example": "Credential definition identifier is not set in proposal",
        },
    )
    revoc_reg_id = fields.Str(
        required=False, metadata={"description": "Revocation registry identifier"}
    )
    revocation_id = fields.Str(
        required=False,
        metadata={"description": "Credential identifier within revocation registry"},
    )



from unittest import IsolatedAsyncioTestCase, mock

from ......connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......utils.testing import create_test_profile
from ......wallet.base import BaseWallet
from ......wallet.did_method import SOV, DIDMethods
from ......wallet.key_type import ED25519
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import DIDX_RESPONSE
from ..response import DIDXResponse


class TestConfig:
    test_seed = "testseed000000000000000000000001"
    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_endpoint = "http://localhost"

    def make_did_doc(self):
        doc = DIDDoc(did=self.test_did)
        controller = self.test_did
        ident = "1"
        pk_value = self.test_verkey
        pk = PublicKey(
            self.test_did,
            ident,
            pk_value,
            PublicKeyType.ED25519_SIG_2018,
            controller,
            False,
        )
        doc.set(pk)
        recip_keys = [pk]
        routing_keys = []
        service = Service(
            self.test_did,
            "indy",
            "IndyAgent",
            recip_keys,
            routing_keys,
            self.test_endpoint,
        )
        doc.set(service)
        return doc


class TestDIDXResponse(IsolatedAsyncioTestCase, TestConfig):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)

            self.did_info = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
            )

            did_doc_attach = AttachDecorator.data_base64(self.make_did_doc().serialize())
            await did_doc_attach.data.sign(self.did_info.verkey, wallet)
            did_rotate_attach = AttachDecorator.data_base64_string(self.test_verkey)
            await did_rotate_attach.data.sign(self.did_info.verkey, wallet)

        self.response = DIDXResponse(
            did=TestConfig.test_did,
            did_doc_attach=did_doc_attach,
            did_rotate_attach=did_rotate_attach,
        )

    def test_init(self):
        """Test initialization."""
        assert self.response.did == TestConfig.test_did

    def test_type(self):
        assert self.response._type == DIDCommPrefix.qualify_current(DIDX_RESPONSE)

    @mock.patch(
        "acapy_agent.protocols.didexchange.v1_0.messages.response.DIDXResponseSchema.load"
    )
    def test_deserialize(self, mock_response_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        response = DIDXResponse.deserialize(obj)
        mock_response_schema_load.assert_called_once_with(obj)

        assert response is mock_response_schema_load.return_value

    @mock.patch(
        "acapy_agent.protocols.didexchange.v1_0.messages.response.DIDXResponseSchema.dump"
    )
    def test_serialize(self, mock_response_schema_dump):
        """
        Test serialization.
        """
        response_dict = self.response.serialize()
        mock_response_schema_dump.assert_called_once_with(self.response)

        assert response_dict is mock_response_schema_dump.return_value


class TestDIDXResponseSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test response schema."""

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)

            self.did_info = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
            )

            did_doc_attach = AttachDecorator.data_base64(self.make_did_doc().serialize())
            await did_doc_attach.data.sign(self.did_info.verkey, wallet)
            did_rotate_attach = AttachDecorator.data_base64_string(self.test_verkey)
            await did_rotate_attach.data.sign(self.did_info.verkey, wallet)

        self.response = DIDXResponse(
            did=TestConfig.test_did,
            did_doc_attach=did_doc_attach,
            did_rotate_attach=did_rotate_attach,
        )

    async def test_make_model(self):
        data = self.response.serialize()
        model_instance = DIDXResponse.deserialize(data)
        assert isinstance(model_instance, DIDXResponse)
        assert model_instance.did_rotate_attach



"""Revocation artifacts."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields, validate

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    BASE58_SHA256_HASH_EXAMPLE,
    BASE58_SHA256_HASH_VALIDATE,
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    NATURAL_NUM_EXAMPLE,
    NATURAL_NUM_VALIDATE,
)


class IndyRevRegDefValuePublicKeysAccumKey(BaseModel):
    """Indy revocation registry definition value public keys accum key."""

    class Meta:
        """Indy revocation registry definition value public keys accum key metadata."""

        schema_class = "IndyRevRegDefValuePublicKeysAccumKeySchema"

    def __init__(self, z: Optional[str] = None):
        """Initialize."""

        self.z = z


class IndyRevRegDefValuePublicKeysAccumKeySchema(BaseModelSchema):
    """Indy revocation registry definition value public keys accum key schema."""

    class Meta:
        """Schema metadata."""

        model_class = IndyRevRegDefValuePublicKeysAccumKey
        unknown = EXCLUDE

    z = fields.Str(
        metadata={
            "description": "Value for z",
            "example": "1 120F522F81E6B7 1 09F7A59005C4939854",
        }
    )


class IndyRevRegDefValuePublicKeys(BaseModel):
    """Indy revocation registry definition value public keys."""

    class Meta:
        """Model metadata."""

        schema_class = "IndyRevRegDefValuePublicKeysSchema"

    def __init__(self, accum_key: Optional[IndyRevRegDefValuePublicKeysAccumKey] = None):
        """Initialize."""

        self.accum_key = accum_key


class IndyRevRegDefValuePublicKeysSchema(BaseModelSchema):
    """Indy revocation registry definition value public keys schema."""

    class Meta:
        """Schema metadata."""

        model_class = IndyRevRegDefValuePublicKeys
        unknown = EXCLUDE

    accum_key = fields.Nested(
        IndyRevRegDefValuePublicKeysAccumKeySchema(), data_key="accumKey"
    )


class IndyRevRegDefValue(BaseModel):
    """Indy revocation registry definition value."""

    class Meta:
        """Model metadata."""

        schema_class = "IndyRevRegDefValueSchema"

    def __init__(
        self,
        issuance_type: Optional[str] = None,
        max_cred_num: Optional[int] = None,
        public_keys: Optional[IndyRevRegDefValuePublicKeys] = None,
        tails_hash: Optional[str] = None,
        tails_location: Optional[str] = None,
    ):
        """Initialize."""
        self.issuance_type = issuance_type
        self.max_cred_num = max_cred_num
        self.public_keys = public_keys
        self.tails_hash = tails_hash
        self.tails_location = tails_location


class IndyRevRegDefValueSchema(BaseModelSchema):
    """Indy revocation registry definition value schema."""

    class Meta:
        """Schema metadata."""

        model_class = IndyRevRegDefValue
        unknown = EXCLUDE

    issuance_type = fields.Str(
        validate=validate.OneOf(["ISSUANCE_ON_DEMAND", "ISSUANCE_BY_DEFAULT"]),
        data_key="issuanceType",
        metadata={"description": "Issuance type"},
    )
    max_cred_num = fields.Int(
        data_key="maxCredNum",
        validate=NATURAL_NUM_VALIDATE,
        metadata={
            "description": "Maximum number of credentials; registry size",
            "strict": True,
            "example": NATURAL_NUM_EXAMPLE,
        },
    )
    public_keys = fields.Nested(
        IndyRevRegDefValuePublicKeysSchema(),
        data_key="publicKeys",
        metadata={"description": "Public keys"},
    )
    tails_hash = fields.Str(
        data_key="tailsHash",
        validate=BASE58_SHA256_HASH_VALIDATE,
        metadata={
            "description": "Tails hash value",
            "example": BASE58_SHA256_HASH_EXAMPLE,
        },
    )
    tails_location = fields.Str(
        data_key="tailsLocation", metadata={"description": "Tails file location"}
    )


class IndyRevRegDef(BaseModel):
    """Indy revocation registry definition."""

    class Meta:
        """Model metadata."""

        schema_class = "IndyRevRegDefSchema"

    def __init__(
        self,
        ver: Optional[str] = None,
        id_: Optional[str] = None,
        revoc_def_type: Optional[str] = None,
        tag: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        value: Optional[IndyRevRegDefValue] = None,
    ):
        """Initialize."""

        self.ver = ver
        self.id_ = id_
        self.revoc_def_type = revoc_def_type
        self.tag = tag
        self.cred_def_id = cred_def_id
        self.value = value


class IndyRevRegDefSchema(BaseModelSchema):
    """Indy revocation registry definition schema."""

    class Meta:
        """Schema metadata."""

        model_class = IndyRevRegDef
        unknown = EXCLUDE

    ver = fields.Str(
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Version of revocation registry definition",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    id_ = fields.Str(
        data_key="id",
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Indy revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    revoc_def_type = fields.Str(
        data_key="revocDefType",
        validate=validate.Equal("CL_ACCUM"),
        metadata={
            "description": "Revocation registry type (specify CL_ACCUM)",
            "example": "CL_ACCUM",
        },
    )
    tag = fields.Str(metadata={"description": "Revocation registry tag"})
    cred_def_id = fields.Str(
        data_key="credDefId",
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    value = fields.Nested(
        IndyRevRegDefValueSchema(),
        metadata={"description": "Revocation registry definition value"},
    )


class IndyRevRegEntryValue(BaseModel):
    """Indy revocation registry entry value."""

    class Meta:
        """Model metadata."""

        schema_class = "IndyRevRegEntryValueSchema"

    def __init__(
        self,
        prev_accum: Optional[str] = None,
        accum: Optional[str] = None,
        revoked: Sequence[int] = None,
    ):
        """Initialize."""
        self.prev_accum = prev_accum
        self.accum = accum
        self.revoked = revoked


class IndyRevRegEntryValueSchema(BaseModelSchema):
    """Indy revocation registry entry value schema."""

    class Meta:
        """Schema metadata."""

        model_class = "IndyRevRegEntryValue"
        unknown = EXCLUDE

    prev_accum = fields.Str(
        data_key="prevAccum",
        required=False,
        metadata={
            "description": "Previous accumulator value",
            "example": "21 137AC810975E4 6 76F0384B6F23",
        },
    )
    accum = fields.Str(
        metadata={
            "description": "Accumulator value",
            "example": "21 11792B036AED0AAA12A4 4 298B2571FFC63A737",
        }
    )
    revoked = fields.List(
        fields.Int(metadata={"strict": True}),
        required=False,
        metadata={"description": "Revoked credential revocation identifiers"},
    )


class IndyRevRegEntry(BaseModel):
    """Indy revocation registry entry."""

    class Meta:
        """Model metadata."""

        schema_class = "IndyRevRegEntrySchema"

    def __init__(
        self, ver: Optional[str] = None, value: Optional[IndyRevRegEntryValue] = None
    ):
        """Initialize."""

        self.ver = ver
        self.value = value


class IndyRevRegEntrySchema(BaseModelSchema):
    """Indy revocation registry entry schema."""

    class Meta:
        """Schema metadata."""

        model_class = IndyRevRegEntry
        unknown = EXCLUDE

    ver = fields.Str(
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Version of revocation registry entry",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    value = fields.Nested(
        IndyRevRegEntryValueSchema(),
        metadata={"description": "Revocation registry entry value"},
    )



import pytest
import pytest_asyncio

from ......core.profile import ProfileSession
from ......utils.testing import create_test_profile
from ...messages.invitation import InvitationMessage
from ..oob_record import OobRecord


@pytest_asyncio.fixture
async def session():
    profile = await create_test_profile()
    async with profile.session() as session:
        yield session


@pytest.mark.asyncio
async def test_oob_record_multi_use(session: ProfileSession):
    """Test oob record multi_use."""
    invi_msg = InvitationMessage(handshake_protocols=["connections/1.0"])
    oob_rec = OobRecord(
        state=OobRecord.STATE_INITIAL,
        invi_msg_id="67890",
        role=OobRecord.ROLE_SENDER,
        invitation=invi_msg,
        multi_use=True,
    )

    await oob_rec.save(session)
    saved_record = await OobRecord.retrieve_by_id(session, record_id=oob_rec.oob_id)
    assert saved_record
    assert isinstance(saved_record, OobRecord)
    assert oob_rec.multi_use
    assert saved_record.multi_use



from ..config.injection_context import InjectionContext
from ..config.provider import ClassProvider
from ..core.profile import Profile


async def setup(context: InjectionContext):
    """Setup vc plugin."""
    from .vc_ld.manager import VcLdpManager

    context.injector.bind_provider(
        VcLdpManager, ClassProvider(VcLdpManager, ClassProvider.Inject(Profile))
    )



"""Abstract wire format classes."""

import json
import logging
from abc import abstractmethod
from typing import List, Sequence, Tuple, Union

from ..core.profile import ProfileSession
from ..messaging.util import time_now
from .error import WireFormatParseError
from .inbound.receipt import MessageReceipt

LOGGER = logging.getLogger(__name__)

DIDCOMM_V0_MIME_TYPE = "application/ssi-agent-wire"
DIDCOMM_V1_MIME_TYPE = "application/didcomm-envelope-enc"


class BaseWireFormat:
    """Abstract messaging wire format."""

    def __init__(self):
        """Initialize the base wire format instance."""

    @abstractmethod
    async def parse_message(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
    ) -> Tuple[dict, MessageReceipt]:
        """Deserialize an incoming message and further populate the request context.

        Args:
            session: The profile session for providing wallet access
            message_body: The body of the message

        Returns:
            A tuple of the parsed message and a message receipt instance

        Raises:
            WireFormatParseError: If the message can't be parsed

        """

    @abstractmethod
    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """Encode an outgoing message for transport.

        Args:
            session: The profile session for providing wallet access
            message_json: The message body to serialize
            recipient_keys: A sequence of recipient verkeys
            routing_keys: A sequence of routing verkeys
            sender_key: The verification key of the sending agent

        Returns:
            The encoded message

        Raises:
            MessageEncodeError: If the message could not be encoded

        """

    @abstractmethod
    def get_recipient_keys(self, message_body: Union[str, bytes]) -> List[str]:
        """Get all recipient keys from a wire message.

        Args:
            message_body: The body of the message

        Returns:
            List of recipient keys from the message body

        Raises:
            RecipientKeysError: If the recipient keys could not be extracted

        """


class JsonWireFormat(BaseWireFormat):
    """Unencrypted wire format."""

    @abstractmethod
    async def parse_message(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
    ) -> Tuple[dict, MessageReceipt]:
        """Deserialize an incoming message and further populate the request context.

        Args:
            session: The profile session for providing wallet access
            message_body: The body of the message

        Returns:
            A tuple of the parsed message and a message receipt instance

        Raises:
            WireFormatParseError: If the JSON parsing failed

        """
        receipt = MessageReceipt()
        receipt.in_time = time_now()
        receipt.raw_message = message_body

        message_dict = None
        message_json = message_body

        if not message_json:
            raise WireFormatParseError("Message body is empty")

        try:
            message_dict = json.loads(message_json)
        except ValueError:
            raise WireFormatParseError("Message JSON parsing failed")
        if not isinstance(message_dict, dict):
            raise WireFormatParseError("Message JSON result is not an object")

        # parse thread ID
        thread_dec = message_dict.get("~thread")
        receipt.thread_id = (
            thread_dec and thread_dec.get("thid") or message_dict.get("@id")
        )
        receipt.parent_thread_id = thread_dec.get("pthid") if thread_dec else None

        # handle transport decorator
        transport_dec = message_dict.get("~transport")
        if transport_dec:
            receipt.direct_response_mode = transport_dec.get("return_route")

        LOGGER.debug(f"Expanded message: {message_dict}")

        return message_dict, receipt

    @abstractmethod
    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """Encode an outgoing message for transport.

        Args:
            session: The profile session for providing wallet access
            message_json: The message body to serialize
            recipient_keys: A sequence of recipient verkeys
            routing_keys: A sequence of routing verkeys
            sender_key: The verification key of the sending agent

        Returns:
            The encoded message

        Raises:
            MessageEncodeError: If the message could not be encoded

        """
        return message_json

    def get_recipient_keys(self, message_body: Union[str, bytes]) -> List[str]:
        """Get all recipient keys from a wire message.

        Args:
            message_body: The body of the message

        Returns:
            List of recipient keys from the message body

        Raises:
            RecipientKeysError: If the recipient keys could not be extracted

        """

        # JSON message cannot contain recipient keys
        return []



import gc
import json
from typing import Optional
from unittest import IsolatedAsyncioTestCase

import jwt
import pytest
import pytest_asyncio
from aiohttp import ClientSession, DummyCookieJar, TCPConnector, web
from aiohttp.test_utils import unused_port
from marshmallow import ValidationError

from ...askar.profile import AskarProfile
from ...config.default_context import DefaultContextBuilder
from ...config.injection_context import InjectionContext
from ...core.event_bus import Event
from ...core.goal_code_registry import GoalCodeRegistry
from ...core.protocol_registry import ProtocolRegistry
from ...multitenant.error import MultitenantManagerError
from ...storage.base import BaseStorage
from ...storage.error import StorageNotFoundError
from ...storage.record import StorageRecord
from ...storage.type import RECORD_TYPE_ACAPY_UPGRADING
from ...tests import mock
from ...utils.stats import Collector
from ...utils.task_queue import TaskQueue
from ...utils.testing import create_test_profile
from ...wallet import singletons
from ...wallet.anoncreds_upgrade import UPGRADING_RECORD_IN_PROGRESS
from .. import server as test_module
from ..request_context import AdminRequestContext
from ..server import AdminServer, AdminSetupError


# Ignore Marshmallow warning, as well as 'NotAppKeyWarning' coming from apispec packages
@pytest.mark.filterwarnings(
    "ignore:The 'missing' attribute of fields is deprecated. Use 'load_default' instead.",
    "ignore:It is recommended to use web.AppKey instances for keys.",
)
class TestAdminServer(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.message_results = []
        self.webhook_results = []
        self.port = 0

        self.connector = TCPConnector(limit=16, limit_per_host=4)
        self.client_session = ClientSession(
            cookie_jar=DummyCookieJar(), connector=self.connector
        )

    async def asyncTearDown(self):
        if self.client_session:
            await self.client_session.close()
        if self.connector:
            await self.connector.close()

    async def test_debug_middleware(self):
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.isEnabledFor = mock.MagicMock(return_value=True)
            mock_logger.debug = mock.MagicMock()

            request = mock.MagicMock(
                method="GET",
                path_qs="/hello/world?a=1&b=2",
                match_info={"match": "info"},
                text=mock.CoroutineMock(return_value="abc123"),
            )
            handler = mock.CoroutineMock()

            await test_module.debug_middleware(request, handler)
            mock_logger.isEnabledFor.assert_called_once()
            assert mock_logger.debug.call_count == 3

    async def test_ready_middleware(self):
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.isEnabledFor = mock.MagicMock(return_value=True)
            mock_logger.debug = mock.MagicMock()
            mock_logger.info = mock.MagicMock()
            mock_logger.error = mock.MagicMock()

            request = mock.MagicMock(
                rel_url="/", app=mock.MagicMock(_state={"ready": False})
            )
            handler = mock.CoroutineMock(return_value="OK")
            with self.assertRaises(test_module.web.HTTPServiceUnavailable):
                await test_module.ready_middleware(request, handler)

            request.app._state["ready"] = True
            assert await test_module.ready_middleware(request, handler) == "OK"

            request.app._state["ready"] = True
            handler = mock.CoroutineMock(
                side_effect=test_module.LedgerConfigError("Bad config")
            )
            with self.assertRaises(test_module.LedgerConfigError):
                await test_module.ready_middleware(request, handler)

            request.app._state["ready"] = True
            handler = mock.CoroutineMock(
                side_effect=test_module.web.HTTPFound(location="/api/doc")
            )
            with self.assertRaises(test_module.web.HTTPFound):
                await test_module.ready_middleware(request, handler)

            request.app._state["ready"] = True
            handler = mock.CoroutineMock(
                side_effect=test_module.asyncio.CancelledError("Cancelled")
            )
            with self.assertRaises(test_module.asyncio.CancelledError):
                await test_module.ready_middleware(request, handler)

            request.app._state["ready"] = True
            handler = mock.CoroutineMock(side_effect=KeyError("No such thing"))
            with self.assertRaises(KeyError):
                await test_module.ready_middleware(request, handler)

    async def test_ready_middleware_http_unauthorized(self):
        """Test handling of web.HTTPUnauthorized and related exceptions."""
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.info = mock.MagicMock()

            request = mock.MagicMock(
                method="GET",
                path="/unauthorized",
                app=mock.MagicMock(_state={"ready": True}),
            )

            # Test web.HTTPUnauthorized
            handler = mock.CoroutineMock(
                side_effect=web.HTTPUnauthorized(reason="Unauthorized")
            )
            with self.assertRaises(web.HTTPUnauthorized):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Unauthorized access during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

            # Test jwt.InvalidTokenError
            handler = mock.CoroutineMock(
                side_effect=jwt.InvalidTokenError("Invalid token")
            )
            with self.assertRaises(web.HTTPUnauthorized):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Unauthorized access during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

            # Test InvalidTokenError
            handler = mock.CoroutineMock(
                side_effect=test_module.InvalidTokenError("Token error")
            )
            with self.assertRaises(web.HTTPUnauthorized):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Unauthorized access during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

    async def test_ready_middleware_http_bad_request(self):
        """Test handling of web.HTTPBadRequest and MultitenantManagerError."""
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.info = mock.MagicMock()

            request = mock.MagicMock(
                method="POST",
                path="/bad-request",
                app=mock.MagicMock(_state={"ready": True}),
            )

            # Test web.HTTPBadRequest
            handler = mock.CoroutineMock(
                side_effect=web.HTTPBadRequest(reason="Bad request")
            )
            with self.assertRaises(web.HTTPBadRequest):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Bad request during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

            # Test MultitenantManagerError
            handler = mock.CoroutineMock(
                side_effect=MultitenantManagerError("Multitenant error")
            )
            with self.assertRaises(web.HTTPBadRequest):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Bad request during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

    async def test_ready_middleware_http_not_found(self):
        """Test handling of web.HTTPNotFound and StorageNotFoundError."""
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.info = mock.MagicMock()

            request = mock.MagicMock(
                method="GET",
                path="/not-found",
                app=mock.MagicMock(_state={"ready": True}),
            )

            # Test web.HTTPNotFound
            handler = mock.CoroutineMock(side_effect=web.HTTPNotFound(reason="Not found"))
            with self.assertRaises(web.HTTPNotFound):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Not Found error occurred during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

            # Test StorageNotFoundError
            handler = mock.CoroutineMock(
                side_effect=StorageNotFoundError("Item not found")
            )
            with self.assertRaises(web.HTTPNotFound):
                await test_module.ready_middleware(request, handler)
            mock_logger.info.assert_called_with(
                "Not Found error occurred during %s %s: %s",
                request.method,
                request.path,
                handler.side_effect,
            )

    async def test_ready_middleware_http_unprocessable_entity(self):
        """Test handling of web.HTTPUnprocessableEntity with nested ValidationError."""
        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            mock_logger.info = mock.MagicMock()
            # Mock the extract_validation_error_message function
            with mock.patch.object(
                test_module, "extract_validation_error_message"
            ) as mock_extract:
                mock_extract.return_value = {"field": ["Invalid input"]}

                request = mock.MagicMock(
                    method="POST",
                    path="/unprocessable",
                    app=mock.MagicMock(_state={"ready": True}),
                )

                # Create a HTTPUnprocessableEntity exception with a nested ValidationError
                validation_error = ValidationError({"field": ["Invalid input"]})
                http_error = web.HTTPUnprocessableEntity(reason="Unprocessable Entity")
                http_error.__cause__ = validation_error

                handler = mock.CoroutineMock(side_effect=http_error)
                with self.assertRaises(web.HTTPUnprocessableEntity):
                    await test_module.ready_middleware(request, handler)
                mock_extract.assert_called_once_with(http_error)
                mock_logger.info.assert_called_with(
                    "Unprocessable Entity occurred during %s %s: %s",
                    request.method,
                    request.path,
                    mock_extract.return_value,
                )

    async def get_admin_server(
        self, settings: Optional[dict] = None, context: Optional[InjectionContext] = None
    ) -> AdminServer:
        if not context:
            context = InjectionContext()
        if settings:
            context.update_settings(settings)

        # middleware is task queue xor collector: cover both over test suite
        task_queue = (settings or {}).pop("task_queue", None)

        plugin_registry = mock.MagicMock(test_module.PluginRegistry, autospec=True)
        plugin_registry.post_process_routes = mock.MagicMock()
        context.injector.bind_instance(test_module.PluginRegistry, plugin_registry)
        context.injector.bind_instance(test_module.Collector, Collector())

        self.profile = await create_test_profile(settings=settings)

        self.port = unused_port()
        return AdminServer(
            "0.0.0.0",
            self.port,
            context,
            self.profile,
            self.outbound_message_router,
            self.webhook_router,
            conductor_stop=mock.CoroutineMock(),
            task_queue=TaskQueue(max_active=4) if task_queue else None,
            conductor_stats=(
                None if task_queue else mock.CoroutineMock(return_value={"a": 1})
            ),
        )

    async def outbound_message_router(self, *args):
        self.message_results.append(args)

    def webhook_router(self, *args):
        self.webhook_results.append(args)

    async def test_start_stop(self):
        with self.assertRaises(AssertionError):
            await (await self.get_admin_server()).start()

        settings = {"admin.admin_insecure_mode": False}
        with self.assertRaises(AssertionError):
            await (await self.get_admin_server(settings)).start()

        settings = {
            "admin.admin_insecure_mode": True,
            "admin.admin_api_key": "test-api-key",
        }
        with self.assertRaises(AssertionError):
            await (await self.get_admin_server(settings)).start()

        settings = {
            "admin.admin_insecure_mode": False,
            "admin.admin_client_max_request_size": 4,
            "admin.admin_api_key": "test-api-key",
        }
        server = await self.get_admin_server(settings)
        await server.start()
        assert server.app._client_max_size == 4 * 1024 * 1024
        with mock.patch.object(server, "websocket_queues", mock.MagicMock()) as mock_wsq:
            mock_wsq.values = mock.MagicMock(
                return_value=[mock.MagicMock(stop=mock.MagicMock())]
            )
            await server.stop()

        with mock.patch.object(web.TCPSite, "start", mock.CoroutineMock()) as mock_start:
            mock_start.side_effect = OSError("Failure to launch")
            with self.assertRaises(AdminSetupError):
                await (await self.get_admin_server(settings)).start()

    async def test_import_routes(self):
        # this test just imports all default admin routes
        # for routes with associated tests, this shouldn't make a difference in coverage
        context = InjectionContext()
        context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())
        context.injector.bind_instance(GoalCodeRegistry, GoalCodeRegistry())
        await DefaultContextBuilder().load_plugins(context)
        server = await self.get_admin_server({"admin.admin_insecure_mode": True}, context)
        await server.make_application()

    async def test_register_external_plugin_x(self):
        context = InjectionContext()
        context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())
        context.injector.bind_instance(GoalCodeRegistry, GoalCodeRegistry())
        with self.assertLogs(level="ERROR") as logs:
            builder = DefaultContextBuilder(
                settings={"external_plugins": ["acapy_agent.nosuchmodule"]}
            )
            await builder.load_plugins(context)
        assert "Module doesn't exist: acapy_agent.nosuchmodule" in "\n".join(logs.output)

    async def test_visit_insecure_mode(self):
        settings = {"admin.admin_insecure_mode": True, "task_queue": True}
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.post(
            f"http://127.0.0.1:{self.port}/status/reset", headers={}
        ) as response:
            assert response.status == 200

        async with self.client_session.ws_connect(
            f"http://127.0.0.1:{self.port}/ws"
        ) as ws:
            result = await ws.receive_json()
            assert result["topic"] == "settings"

        for path in (
            "",
            "plugins",
            "status",
            "status/live",
            "status/ready",
            "shutdown",  # mock conductor has magic-mock stop()
        ):
            async with self.client_session.get(
                f"http://127.0.0.1:{self.port}/{path}", headers={}
            ) as response:
                assert response.status == 200

        await server.stop()

    async def test_visit_secure_mode(self):
        settings = {
            "admin.admin_insecure_mode": False,
            "admin.admin_api_key": "test-api-key",
        }
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status", headers={"x-api-key": "wrong-key"}
        ) as response:
            assert response.status == 401

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status",
            headers={"x-api-key": "test-api-key"},
        ) as response:
            assert response.status == 200

        # Make sure that OPTIONS requests used by browsers for CORS
        # are allowed without a x-api-key even when x-api-key security is enabled
        async with self.client_session.options(
            f"http://127.0.0.1:{self.port}/status",
            headers={
                "Access-Control-Request-Headers": "x-api-key",
                "Access-Control-Request-Method": "GET",
                "Connection": "keep-alive",
                "Host": f"http://127.0.0.1:{self.port}/status",
                "Origin": "http://localhost:3000",
                "Referer": "http://localhost:3000/",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-site",
            },
        ) as response:
            assert response.status == 200
            assert response.headers["Access-Control-Allow-Credentials"] == "true"
            assert response.headers["Access-Control-Allow-Headers"] == "X-API-KEY"
            assert response.headers["Access-Control-Allow-Methods"] == "GET"
            assert (
                response.headers["Access-Control-Allow-Origin"] == "http://localhost:3000"
            )

        async with self.client_session.ws_connect(
            f"http://127.0.0.1:{self.port}/ws", headers={"x-api-key": "test-api-key"}
        ) as ws:
            result = await ws.receive_json()
            assert result["topic"] == "settings"

        await server.stop()

    async def test_query_config(self):
        settings = {
            "admin.admin_insecure_mode": False,
            "admin.admin_api_key": "test-api-key",
            "admin.webhook_urls": ["localhost:8123/abc#secret", "localhost:8123/def"],
            "multitenant.jwt_secret": "abc123",
            "wallet.key": "abc123",
            "wallet.rekey": "def456",
            "wallet.seed": "00000000000000000000000000000000",
            "wallet.storage.creds": "secret",
        }
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/config",
            headers={"x-api-key": "test-api-key"},
        ) as response:
            config = json.loads(await response.text())["config"]
            assert "admin.admin_insecure_mode" in config
            assert all(
                k not in config
                for k in [
                    "admin.admin_api_key",
                    "multitenant.jwt_secret",
                    "wallet.key",
                    "wallet.rekey",
                    "wallet.seed",
                    "wallet.storage_creds",
                ]
            )
            assert config["admin.webhook_urls"] == [
                "localhost:8123/abc",
                "localhost:8123/def",
            ]

    async def test_visit_shutting_down(self):
        settings = {
            "admin.admin_insecure_mode": True,
        }
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/shutdown", headers={}
        ) as response:
            assert response.status == 200

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status", headers={}
        ) as response:
            assert response.status == 503

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/live", headers={}
        ) as response:
            assert response.status == 200
        await server.stop()

    async def test_server_health_state(self):
        settings = {
            "admin.admin_insecure_mode": True,
        }
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/live", headers={}
        ) as response:
            assert response.status == 200
            response_json = await response.json()
            assert response_json["alive"]

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/ready", headers={}
        ) as response:
            assert response.status == 200
            response_json = await response.json()
            assert response_json["ready"]

        server.notify_fatal_error()
        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/live", headers={}
        ) as response:
            assert response.status == 503

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/ready", headers={}
        ) as response:
            assert response.status == 503
        await server.stop()

    async def test_server_aiohttp_headers_removed(self):
        settings = {
            "admin.admin_insecure_mode": True,
        }
        server = await self.get_admin_server(settings)
        await server.start()

        async with self.client_session.get(
            f"http://127.0.0.1:{self.port}/status/live", headers={}
        ) as response:
            assert response.headers.get("Server") is None

        await server.stop()

    async def test_upgrade_middleware(self):
        profile = await create_test_profile()
        self.context = AdminRequestContext.test_context({}, profile)
        self.request_dict = {
            "context": self.context,
        }
        request = mock.MagicMock(
            method="GET",
            path_qs="/schemas/created",
            match_info={},
            __getitem__=lambda _, k: self.request_dict[k],
        )
        handler = mock.CoroutineMock()

        await test_module.upgrade_middleware(request, handler)

        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            upgrading_record = StorageRecord(
                RECORD_TYPE_ACAPY_UPGRADING,
                UPGRADING_RECORD_IN_PROGRESS,
            )
            # No upgrade in progress
            await storage.add_record(upgrading_record)

            # Upgrade in progress without cache
            with self.assertRaises(test_module.web.HTTPServiceUnavailable):
                await test_module.upgrade_middleware(request, handler)

            # Upgrade in progress with cache
            singletons.UpgradeInProgressSingleton().set_wallet(profile.name)
            with self.assertRaises(test_module.web.HTTPServiceUnavailable):
                await test_module.upgrade_middleware(request, handler)

            singletons.UpgradeInProgressSingleton().remove_wallet(profile.name)
            await storage.delete_record(upgrading_record)

            # Upgrade in progress with cache
            singletons.IsAnonCredsSingleton().set_wallet(profile.name)
            await test_module.upgrade_middleware(request, handler)


@pytest_asyncio.fixture
async def server():
    test_class = TestAdminServer()
    await test_class.asyncSetUp()
    yield test_class.get_admin_server()
    await test_class.asyncTearDown()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "event_topic, webhook_topic",
    [("acapy::record::topic", "topic"), ("acapy::record::topic::state", "topic")],
)
async def test_on_record_event(server, event_topic, webhook_topic):
    profile = mock.MagicMock(AskarProfile, autospec=True)
    server = await server
    with mock.patch.object(
        server, "send_webhook", mock.CoroutineMock()
    ) as mock_send_webhook:
        await server._on_record_event(profile, Event(event_topic, None))
        mock_send_webhook.assert_called_once_with(profile, webhook_topic, None)


@pytest.mark.asyncio
async def test_admin_responder_profile_expired_x():
    async def _smaller_scope():
        profile = await create_test_profile()
        return test_module.AdminResponder(profile, None)

    responder = await _smaller_scope()
    gc.collect()  # help ensure collection of profile

    with pytest.raises(RuntimeError):
        await responder.send_outbound(None)

    with pytest.deprecated_call():
        with pytest.raises(RuntimeError):
            await responder.send_webhook("test", {})
